//Required libraries
#include <Preferences.h>
#include <Adafruit_MPU6050.h>
#include <Adafruit_Sensor.h>
#include <Wire.h>

Adafruit_MPU6050 mpu;

//Pin defintions
#define HLF         34 //Height sense left front
#define HLB         35 //Height sense left back
#define HRF         32 //Height sense right front
#define HRB         33 //Height sense right back
#define MLF1        18 //Motor left front 1
#define MLF2        19 //Motor left front 2
#define MLB1        26 //Motor left back 1
#define MLB2        25 //Motor left back 2
#define MRF1        15 //Motor right front 1
#define MRF2        14 //Motor right front 2
#define MRB1        23 //Motor right back 1
#define MRB2        4  //Motor right back 2

//Variable definitions
const int freq       = 20000; //PWM frequency: 20 kHz
const int resolution = 8; //PWM with 8 bit
const int PWM_LF1    = 0;
const int PWM_LF2    = 1;
const int PWM_LB1    = 2;
const int PWM_LB2    = 3;
const int PWM_RF1    = 4;
const int PWM_RF2    = 5;
const int PWM_RB1    = 6;
const int PWM_RB2    = 7;

int inval         = 0;
int lfp           = 0;
int lbp           = 0;
int rfp           = 0;
int rbp           = 0;
int runvar        = 0;
int time0         = 0;
int time1         = 0;
int timer         = 0;
int timer1        = 0;
float HLFv        = 0;
float HLBv        = 0;
float HRFv        = 0;
float HRBv        = 0;
float lf          = 0;
float lb          = 0;
float rf          = 0;
float rb          = 0;
float hc          = 0;
int defaultheight = 0;
float xval        = 0;
float yval        = 0;
int motorvar      = 0;
int corr          = 1;
int lim           = 0;
int hcold         = 0;
float thres       = 0.03;
float adj         = 0.1;

Preferences preferences; //used to store settings and values that were adjusted by the remote control and forwarded by the primary µC

void setup() {
  preferences.begin("Defaults", false);
  inval         = preferences.getInt("Height", 30);
  HLFv          = preferences.getInt("leftfront", 30);
  HLBv          = preferences.getInt("leftback", 30);
  HRFv          = preferences.getInt("rightfront", 30);
  HRBv          = preferences.getInt("rightback", 30);
  defaultheight = preferences.getInt("Mode", 111);
  
  //initialize serial monitor - used for debugging
  Serial.begin(115200);

  //initialize serial2 for communication with primary µC
  Serial2.begin(9600, SERIAL_8N1, 16, 17);

  mpu.begin(); //MPU6050 initialization according to the library
  mpu.setAccelerometerRange(MPU6050_RANGE_8_G);
  mpu.setGyroRange(MPU6050_RANGE_500_DEG);
  mpu.setFilterBandwidth(MPU6050_BAND_21_HZ);
  
  //PWM channels for the linear actuators (two channels per motor: 8 in total)
  ledcSetup(PWM_LF1, freq, resolution); //PWM channel, frequency, resolution
  ledcAttachPin(MLF1, PWM_LF1); //Pin, PWM channel
  ledcWrite(PWM_LF1, 0); //PWM channel, default value
  
  ledcSetup(PWM_LF2, freq, resolution); 
  ledcAttachPin(MLF2, PWM_LF2); 
  ledcWrite(PWM_LF2, 0); 

  ledcSetup(PWM_LB1, freq, resolution); 
  ledcAttachPin(MLB1, PWM_LB1); 
  ledcWrite(PWM_LB1, 0);  

  ledcSetup(PWM_LB2, freq, resolution); 
  ledcAttachPin(MLB2, PWM_LB2); 
  ledcWrite(PWM_LB2, 0); 

  ledcSetup(PWM_RF1, freq, resolution); 
  ledcAttachPin(MRF1, PWM_RF1); 
  ledcWrite(PWM_RF1, 0); 
  
  ledcSetup(PWM_RF2, freq, resolution); 
  ledcAttachPin(MRF2, PWM_RF2); 
  ledcWrite(PWM_RF2, 0); 

  ledcSetup(PWM_RB1, freq, resolution); 
  ledcAttachPin(MRB1, PWM_RB1); 
  ledcWrite(PWM_RB1, 0); 

  ledcSetup(PWM_RB2, freq, resolution); 
  ledcAttachPin(MRB2, PWM_RB2); 
  ledcWrite(PWM_RB2, 0); 

  delay(5000); //relatively long wait interval to grant the primary µC enough time to enter the main loop
  time0 = millis();
}

void loop() {
  while(Serial2.available()) {
    inval = Serial2.read();
    if (inval>0&&inval<100){
      preferences.putInt("Height", inval); //save last height value as a preference
    }
  }
  
  if (inval==110){ //change the height adjust mode according to the input from the serial monitor...
    defaultheight = inval;
    preferences.putInt("Mode", defaultheight); //...and save it as a preference
    lf = 0; //a mode change always triggers the tilt to be reset
    lb = 0;
    rf = 0;
    rb = 0;
  }else if (inval==111){
    defaultheight = inval;
    preferences.putInt("Mode", defaultheight);
    lf = 0;
    lb = 0;
    rf = 0;
    rb = 0;
  }

  if (inval==255){ //the actuators stop instantly when this value is transmitted via the serial connection
    MotorStop();
    motorvar = 0;  
  }else if(inval!=110&&inval!=111){    
    HeightSense(); //reads the current heights of the actuators; requires approx. 5 ms!
    if (defaultheight==111){
      HeightUpdate(); //this function changes the tilt values (lf, lb, rf, rb) according to inputs from the primary µC
      motorvar = 0;
      if ((HLFv+lf<hc-1.5&&HLFv<55)||(HLFv+lf>hc+1.5&&HLFv>15)||(HLBv+lb<hc-1.5&&HLBv<55)||(HLBv+lb>hc+1.5&&HLBv>15)||(HRFv+rf<hc-1.5&&HRFv<55)||(HRFv+rf>hc+1.5&&HRFv>15)||(HRBv+rb<hc-1.5&&HRBv<55)||(HRBv+rb>hc+1.5&&HRBv>15)){
        //thresholds for starting an adjustment (two point controllers: on when exceeding limits; off when within certain limits
        //the correction is either on or off for all four actuators at the same time
        //the defined limits are maximum and minimum constraints (55 and 15) and offsets from the center height (hc) that exceed the value of 1.5
        //maintaining tilt is enabled by adding the current tilt value (lf, lb, rf, rb) to each of the four height values
        corr = 1; 
      }else if ((abs(hc-HLFv-lf<0.3)||HLFv>54||HLFv<16)&&(abs(hc-HLBv-lb)<0.3||HLBv>54||HLBv<16)&&(abs(hc-HRFv-rf<0.3)||HRFv>54||HRFv<16)&&(abs(hc-HRBv-rb)<0.3||HRBv>54||HRBv<16)){ 
        //thresholds to stop correction: within the absolute limits (<16 or >54) or within +-0.3 of the desired height value
        corr = 0;
      }      
    }else if (defaultheight==110){ //automatic tilt compensation
      AutoHeight();
    }
    
    if (corr==1){
      MotorUpdate();
      if (motorvar>250&&defaultheight==110){ //endless compensation in automatic tilt adjustment is prevented by the motorvar threshold
        corr = 0; 
      }
    }else{ //stop all motors if no correction is needed
    MotorStop();    
    }
  }
  runvar=runvar+1;

  //the following block used for debugging - note that this code is runtime sensitive: a serial connection to the PC slows down the loop speed. Thus, it is only performed twice per second
  //time1 = millis();
  //if (time1>time0+500){
    //updateSerial(); 
    //time0 = millis();
  //}  
}

//this function reads the MPU6050 and sets the tilt compensation accordingly
void AutoHeight(){
  sensors_event_t a, g, temp;
  mpu.getEvent(&a, &g, &temp);
  xval = a.acceleration.x+0.000; //these offsets were determined manually: use a spirit level and compensate the two orthogonal main axes (x and y) so that they both say 0 when the tank is levelled
  yval = a.acceleration.y+0.075;
  for (int i=0; i<5; i++){ //the usual denoising by averaging the values of several sensor readings
    sensors_event_t a, g, temp;
    mpu.getEvent(&a, &g, &temp);
    xval = xval + a.acceleration.x+0.000;
    yval = yval + a.acceleration.y+0.075;  
  }
  xval = xval/6;
  yval = yval/6;
  
  if (inval>0&&inval<61){ //the height values between 15 and 55 determine the center height of the system
    hc = inval;
    if (hcold!=hc){ //trigger actuator movement if the center height was changed
      corr     = 1;
      motorvar = 0;
    }
    hcold = hc;
  }   
   
  if(hc<25){ //auto tilt has tighter upper and lower boundaries than manual tilt to provide full tilt compensation throughout the whole center height range
    hc=25;
  }else if(hc>45){
    hc=45;
  }    
  
  if (abs(xval)>0.15||abs(yval)>0.15||(abs(((HLFv+HLBv+HRFv+HRBv)/4)-hc)>1.5)){ //compensation is triggered when at least 0.15 G are detected...
    if (motorvar<250){ //...but with a fixed timer limit that prevents endless compensation attempts when reaching limits that prevent a completion of tilt compensation
      corr = 1;    
    }
  }else if (abs(xval)<0.03&&abs(yval)<0.03&&(abs(((HLFv+HLBv+HRFv+HRBv)/4)-hc)<1)){ //compensation stops when the deviations are below these thresholds
    corr = 0;
  }

  if (corr==1){ //specify the correction offsets depending on the current tilt. These offsets are changed incrementally (a little bit on every loop iteration) to prevent overshooting
    if (xval>=+thres){
      lf = lf-adj; //adj equals the increment by which the tilt offset is changed per iteration
      lb = lb-adj;
      rf = rf+adj;
      rb = rb+adj;     
    }else if (xval<=-thres){
      lf = lf+adj;
      lb = lb+adj;
      rf = rf-adj;
      rb = rb-adj;        
    }
    if (yval<=-thres){
      lf = lf-adj;
      lb = lb+adj;
      rf = rf-adj;
      rb = rb+adj;        
    }else if (yval>=thres){
      lf = lf+adj;
      lb = lb-adj;
      rf = rf+adj;
      rb = rb-adj;        
    }
  }
  
  //absolute offset limits: +-15  
  if(lf>=15){
    lf = 15;
  }else if(lf<=-15){
    lf = -15;
  }
  if(lb>=15){
    lb = 15;
  }else if(lb<=-15){
    lb = -15;
  }
  if(rf>=15){
    rf = 15;
  }else if(rf<=-15){
    rf = -15;
  }
  if(rb>=15){
    rb = 15;
  }else if(rb<=-15){
    rb = -15;
  } 

  //offset limit per side: 20  
  if(abs(lf-lb)>=20){
    if(lf>=10){
      lf = 10;
    }else if(lf<=-10){
      lf = -10;
    }
    if(lb>=10){
      lb = 10;
    }else if(lb<=-10){
      lb = -10;
    }
  } 
  if(abs(rf-rb)>=20){
    if(rf>=10){
      rf = 10;
    }else if(rf<=-10){
      rf = -10;
    }
    if(rb>=10){
      rb = 10;
    }else if(rb<=-10){
      rb = -10;
    }
  }      
}

//simple function: instantly stop all motors
void MotorStop(){
  ledcWrite(PWM_LF1,0);
  ledcWrite(PWM_LF2,0); 
  ledcWrite(PWM_LB1,0);
  ledcWrite(PWM_LB2,0);
  ledcWrite(PWM_RF1,0);
  ledcWrite(PWM_RF2,0); 
  ledcWrite(PWM_RB1,0);
  ledcWrite(PWM_RB2,0); 
}

//motor update is called during manual and automatic tilt adjustment: it runs the motors according to which corrections are currently needed
void MotorUpdate(){
  if (HLFv+lf<hc-0.5&&HLFv<55){ //left front up
    ledcWrite(PWM_LF2,200);
    ledcWrite(PWM_LF1,0);
  }else if (HLFv+lf>hc+0.5&&HLFv>15){ //left front down
    ledcWrite(PWM_LF1,200); 
    ledcWrite(PWM_LF2,0);
  }else{ //left front no movement
    ledcWrite(PWM_LF1,0);
    ledcWrite(PWM_LF2,0);
    if (HLFv-lf>54||HLFv-lf<16){ //count up motor timer when this actuator is at the upper or lower limit
      motorvar = motorvar+2;
    }
  }
  if (HLBv+lb<hc-0.5&&HLFv<55){ //left back up
    ledcWrite(PWM_LB2,200);
    ledcWrite(PWM_LB1,0);
  }else if (HLBv+lb>hc+0.5&&HLBv>15){ //left back down
    ledcWrite(PWM_LB1,200);
    ledcWrite(PWM_LB2,0);
  }else{ //left back no movement
    ledcWrite(PWM_LB1,0);
    ledcWrite(PWM_LB2,0);
    if (HLBv-lb>54||HLBv-lb<16){ //count up motor timer when this actuator is at the upper or lower limit
      motorvar = motorvar+2;
    }
  }
  
  if (HRFv+rf<hc-0.5&&HRFv<55){ //right front up
    ledcWrite(PWM_RF2,200);
    ledcWrite(PWM_RF1,0);
  }else if (HRFv+rf>hc+0.5&&HRFv>15){ //right front down
    ledcWrite(PWM_RF1,200);
    ledcWrite(PWM_RF2,0);
  }else{ //right front no movement
    ledcWrite(PWM_RF1,0);
    ledcWrite(PWM_RF2,0);
    if (HRFv-rf>54||HRFv-rf<16){ //count up motor timer when this actuator is at the upper or lower limit
      motorvar = motorvar+2;
    }
  }
  if (HRBv+rb<hc-0.5&&HRBv<55){ //right back up
    ledcWrite(PWM_RB2,200);
    ledcWrite(PWM_RB1,0);
  }else if (HRBv+rb>hc+0.5&&HRBv>15){ //right back down
    ledcWrite(PWM_RB1,200);
    ledcWrite(PWM_RB2,0);
  }else{ //right back no movement
    ledcWrite(PWM_RB1,0);
    ledcWrite(PWM_RB2,0);
    if (HRBv-rb>54||HRBv-rb<16){ //count up motor timer when this actuator is at the upper or lower limit
      motorvar = motorvar+2;
    }
  }
}

//height sense: reading of the four potis and compensating their non-linearity
void HeightSense(){
  lfp = analogRead(HLF);
  lbp = analogRead(HLB);
  rfp = analogRead(HRF);
  rbp = analogRead(HRB);
  for (int i=0; i<5; i++){ //averaging several readings to reduce noise
    lfp = lfp + analogRead(HLF);
    lbp = lbp + analogRead(HLB);
    rfp = rfp + analogRead(HRF);
    rbp = rbp + analogRead(HRB);      
  }
  HLFv = float(lfp)/(6*4095); //the code could run faster if the ADC values would not be converted to float type variables
  HLBv = float(lbp)/(6*4095);
  HRFv = float(rfp)/(6*4095);
  HRBv = float(rbp)/(6*4095);
  HLFv = float(292.9*HLFv*HLFv*HLFv*HLFv-649.3*HLFv*HLFv*HLFv+450.3*HLFv*HLFv-29.9*HLFv-3.7)-0.2;    //individual correction function for HLF inputs, converts analog values into height of the actuator in mm (distance from bottom to carriage)
  HLBv = float(-130.6*HLBv*HLBv*HLBv*HLBv+310.8*HLBv*HLBv*HLBv-237.1*HLBv*HLBv-8.8*HLBv+65.8)+0.2;   //individual correction function for HLB inputs, converts analog values into height of the actuator in mm (distance from bottom to carriage)
  HRFv = float(-28.9*HRFv*HRFv*HRFv*HRFv+82.5*HRFv*HRFv*HRFv-57.6*HRFv*HRFv-67.8*HRFv+71.6)+0.2;     //individual correction function for HRF inputs, converts analog values into height of the actuator in mm (distance from bottom to carriage)
  HRBv = float(-106.8*HRBv*HRBv*HRBv*HRBv+225.7*HRBv*HRBv*HRBv-182.4*HRBv*HRBv+143.6*HRBv-15.3)-0.4; //individual correction function for HRB inputs, converts analog values into height of the actuator in mm (distance from bottom to carriage)
  preferences.putInt("leftfront", HLFv);
  preferences.putInt("leftback", HLBv);
  preferences.putInt("rightfront", HRFv);
  preferences.putInt("rightback", HRBv);          
}

//function for manual height update
void HeightUpdate(){
  if (runvar>4){ //the runvar timer delays manual tilt adjustments so that joystick inputs are internally counted up/down slower
    runvar = 0;    
    if (inval>0&&inval<61){ //values below 60 are actual height values
      hc = inval;
    }
    if(hc<15){ //upper and lower center height limits
      hc = 15;
    }else if(hc>55){
      hc = 55;
    }
    if(inval==109){ //109 means resetting the tilt. Is only executed within the manual mode
      lf       = 0;
      lb       = 0;
      rf       = 0;
      rb       = 0;
      motorvar = 0;
    }
    //incremental tilt changes depending on left joystick position:
    if(inval==101){ //101 = tilt to the left
      lf = lf-0.1;
      lb = lb-0.1;
      rf = rf+0.1;
      rb = rb+0.1;
    }else if(inval==102){ //102 = tilt to left-front
      lf = lf-0.1;
      rb = rb+0.1;
      if(lb<0){ //left-back is corrected to center
        lb = lb+0.1;
      }else if(lb>0){
        lb = lb-0.1;
      }
      if(rf<0){ //right-front is corrected to center
        rf = rf+0.1;
      }else if(rf>0){
        rf = rf-0.1;
      }
    }else if(inval==103){ //103 = tilt to front
      lf = lf-0.1;
      rf = rf-0.1;
      lb = lb+0.1;
      rb = rb+0.1;
    }else if(inval==104){ //104 = tilt to right-front
      rf = rf-0.1;
      lb = lb+0.1;
      if(lf<0){ //left-front is corrected to center
        lf = lf+0.1;
      }else if(lf>0){
        lf = lf-0.1;
      }
      if(rb<0){ //right-back is corrected to center
        rb = rb+0.1;
      }else if(rb>0){
        rb = rb-0.1;
      }    
    }else if(inval==105){ //105 = tilt to right
      rf = rf-0.1;
      rb = rb-0.1;
      lf = lf+0.1;
      lb = lb+0.1;
    }else if(inval==106){ //106 = tilt to right-back
      rb = rb-0.1;
      lf = lf+0.1;
      if(lb<0){ //left-back is corrected to center
        lb = lb+0.1;
      }else if(lb>0){
        lb = lb-0.1;
      }
      if(rf<0){ //right-front is corrected to center
        rf = rf+0.1;
      }else if(rf>0){
        rf = rf-0.1;
      }    
    }else if(inval==107){ //107 = tilt to back
      lb = lb-0.1;
      rb = rb-0.1;
      lf = lf+0.1;
      rf = rf+0.1;
    }else if(inval==108){ //108 = tilt to left-back
      lb = lb-0.1;
      rf = rf+0.1;
      if(lf<0){ //left-front is corrected to center
        lf = lf+0.1;
      }else if(lf>0){
        lf = lf-0.1;
      }
      if(rb<0){ //right-back is corrected to center
        rb = rb+0.1;
      }else if(rb>0){
        rb = rb-0.1;
      }     
    }      
  }
  
  //absolute offset limits: +-15  
  if(lf>15){
    lf = 15;
  }else if(lf<-15){
    lf = -15;
  }
  if(lb>15){
    lb = 15;
  }else if(lb<-15){
    lb = -15;
  }
  if(rf>15){
    rf = 15;
  }else if(rf<-15){
    rf = -15;
  }
  if(rb>15){
    rb = 15;
  }else if(rb<-15){
    rb = -15;
  } 

  //offset limit per side: 20  
  if(abs(lf-lb)>20){
    if(lf>10){
      lf = 10;
    }else if(lf<-10){
      lf = -10;
    }
    if(lb>10){
      lb = 10;
    }else if(lb<-10){
      lb = -10;
    }
  } 
  if(abs(rf-rb)>20){
    if(rf>10){
      rf = 10;
    }else if(rf<-10){
      rf = -10;
    }
    if(rb>10){
      rb = 10;
    }else if(rb<-10){
      rb = -10;
    }
  }      
}

void updateSerial(){
//  Serial.print("Acceleration X: ");
//  Serial.print(xval,3);
//  Serial.print(", Y: ");
//  Serial.print(yval,3);
//  //Serial.print(", Z: ");
//  //Serial.print(a.acceleration.z-2.5,2);
//  Serial.println(" m/s^2"); 
//  Serial.println(" ");
//  Serial.print("Correction: ");
//  Serial.println(corr); 
//  Serial.println(" ");
//  Serial.print("Height adjust mode: ");
//  if (defaultheight==110){
//    Serial.println("Auto");
//  }else if (defaultheight==111){
//    Serial.println("Manual");
//  }else{
//    Serial.println("Error");
//  }
//  Serial.println("----------");
//  Serial.print("Serial2 value:  ");
//  Serial.println(inval);
//  Serial.print("Center height:  ");
//  Serial.println(hc);
//  Serial.print("Motor variable: ");
//  Serial.println(motorvar);
//  Serial.println("----------");
//  Serial.print("Factor left front:  ");
//  Serial.println(lf);
//  Serial.print("Factor left back:   ");
//  Serial.println(lb);
//  Serial.print("Factor right front: ");
//  Serial.println(rf);
//  Serial.print("Factor right back:  ");
//  Serial.println(rb);  
//  Serial.println("----------");  
//  Serial.print("Height left front:  ");
//  Serial.println(HLFv);
//  Serial.print("Height left back:   ");
//  Serial.println(HLBv);
//  Serial.print("Height right front: ");
//  Serial.println(HRFv);
//  Serial.print("Height right back:  ");
//  Serial.println(HRBv);
//  Serial.println("----------");    
}
